/*
  SBR bit writing routines
*/
#include <stdlib.h>
#include "sbr.h"
#include "hufftbls.h"

typedef enum {
  SBR_ID_SCE = 1,
  SBR_ID_CPE
} SBR_ELEMENT_TYPE;

int writeNoiseLevelData (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream, int coupling);
int writeEnvelopeData (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream, int coupling);

/*****************************************************************************
    description:  writes bits corresponding to the "synthetic-coding"-extension
*****************************************************************************/
static int writeSyntheticCodingData (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream)
{
  int i;
  int payloadBits = 0;
  payloadBits += WriteBits (hBitStream, sbrEnvData->addHarmonicFlag, 1);

  if (sbrEnvData->addHarmonicFlag) {
    for (i = 0; i < sbrEnvData->noHarmonics; i++) {
      payloadBits += WriteBits (hBitStream, sbrEnvData->addHarmonic[i], 1);
    }
  }
  return payloadBits;
}

int ceil_ln2(int x)
{
  int tmp=-1;
  while((1<<++tmp) < x);
  return(tmp);
}


/*****************************************************************************
    description:  Encode SBR grid information
*****************************************************************************/
static inline int encodeSbrGrid (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream)
{
  int payloadBits = 0;
  int temp;
  payloadBits += WriteBits (hBitStream, 0, SBR_CLA_BITS);
  temp = ceil_ln2(sbrEnvData->bs_num_env);
  payloadBits += WriteBits (hBitStream, temp, SBR_ENV_BITS);
  payloadBits += WriteBits (hBitStream, FREQ_RES_HIGH, SBR_RES_BITS);
  return payloadBits;
}

/*****************************************************************************
    description:  writes bits that describes the direction of the envelopes of a frame
*****************************************************************************/
static inline int encodeSbrDtdf (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream)
{
  int i, payloadBits = 0, noOfNoiseEnvelopes;
  noOfNoiseEnvelopes = (sbrEnvData->noOfEnvelopes > 1) ? 2 : 1;
  for (i = 0; i < sbrEnvData->noOfEnvelopes; ++i) {
    payloadBits += WriteBits (hBitStream, FREQ, SBR_DIR_BITS);
  }
  for (i = 0; i < noOfNoiseEnvelopes; ++i) {
    payloadBits +=  WriteBits (hBitStream, FREQ, SBR_DIR_BITS);
  }
  return payloadBits;
}

/*****************************************************************************
    description:  encodes sbr SCE information
*****************************************************************************/
static int encodeSbrSingleChannelElement (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream)
{
  int payloadBits = 0;
  payloadBits += WriteBits (hBitStream, 0, 1); /* no reserved bits */
  payloadBits += encodeSbrGrid (sbrEnvData, hBitStream);
  payloadBits += encodeSbrDtdf (sbrEnvData, hBitStream);
  {
    int i;
    for (i = 0; i < sbrEnvData->noOfnoisebands; i++) {
      payloadBits += WriteBits (hBitStream, sbrEnvData->sbr_invf_mode_vec[i], SI_SBR_INVF_MODE_BITS);
    }
  }
  payloadBits += writeEnvelopeData (sbrEnvData, hBitStream, 0);
  payloadBits += writeNoiseLevelData (sbrEnvData, hBitStream, 0);
  payloadBits += writeSyntheticCodingData (sbrEnvData,hBitStream);
  payloadBits += WriteBits (hBitStream, 0, SI_SBR_EXTENDED_DATA_BITS);
  return payloadBits;
}


/*****************************************************************************
    description:  encodes sbr CPE information
*****************************************************************************/
static int encodeSbrChannelPairElement (SBR_ENV_DATA *sbrEnvDataLeft, SBR_ENV_DATA *sbrEnvDataRight,
                                 BIT_BUF *hBitStream, int coupling)
{
  int payloadBits = 0;
  int i = 0;
  payloadBits += WriteBits (hBitStream, 0, 1); /* no reserved bits */
  payloadBits += WriteBits (hBitStream, coupling, SI_SBR_COUPLING_BITS);

  if (coupling) {
    payloadBits += encodeSbrGrid (sbrEnvDataLeft, hBitStream);
    payloadBits += encodeSbrDtdf (sbrEnvDataLeft, hBitStream);
    payloadBits += encodeSbrDtdf (sbrEnvDataRight, hBitStream);

    for (i = 0; i < sbrEnvDataLeft->noOfnoisebands; i++) {
      payloadBits += WriteBits (hBitStream, sbrEnvDataLeft->sbr_invf_mode_vec[i], SI_SBR_INVF_MODE_BITS);
    }

    payloadBits += writeEnvelopeData  (sbrEnvDataLeft,  hBitStream,1);
    payloadBits += writeNoiseLevelData (sbrEnvDataLeft,  hBitStream,1);
    payloadBits += writeEnvelopeData  (sbrEnvDataRight, hBitStream,1);
    payloadBits += writeNoiseLevelData (sbrEnvDataRight, hBitStream,1);
    payloadBits += writeSyntheticCodingData (sbrEnvDataLeft,hBitStream);
    payloadBits += writeSyntheticCodingData (sbrEnvDataRight,hBitStream);
  } else { /* no coupling */
    payloadBits += encodeSbrGrid (sbrEnvDataLeft,  hBitStream);
    payloadBits += encodeSbrGrid (sbrEnvDataRight, hBitStream);
    payloadBits += encodeSbrDtdf (sbrEnvDataLeft,  hBitStream);
    payloadBits += encodeSbrDtdf (sbrEnvDataRight, hBitStream);

    for (i = 0; i < sbrEnvDataLeft->noOfnoisebands; i++) {
      payloadBits += WriteBits (hBitStream, sbrEnvDataLeft->sbr_invf_mode_vec[i], SI_SBR_INVF_MODE_BITS);
    }

    for (i = 0; i < sbrEnvDataRight->noOfnoisebands; i++) {
      payloadBits += WriteBits (hBitStream, sbrEnvDataRight->sbr_invf_mode_vec[i], SI_SBR_INVF_MODE_BITS);
    }

    payloadBits += writeEnvelopeData  (sbrEnvDataLeft,  hBitStream,0);
    payloadBits += writeEnvelopeData  (sbrEnvDataRight, hBitStream,0);
    payloadBits += writeNoiseLevelData (sbrEnvDataLeft,  hBitStream,0);
    payloadBits += writeNoiseLevelData (sbrEnvDataRight, hBitStream,0);
    payloadBits += writeSyntheticCodingData (sbrEnvDataLeft,hBitStream);
    payloadBits += writeSyntheticCodingData (sbrEnvDataRight,hBitStream);
  } /* coupling */
  payloadBits += WriteBits (hBitStream, 0, SI_SBR_EXTENDED_DATA_BITS);
  return payloadBits;
}

/*****************************************************************************
    description:  encodes sbr Data information
*****************************************************************************/
static int encodeSbrData (SBR_ENV_DATA *sbrEnvDataLeft, SBR_ENV_DATA *sbrEnvDataRight,
                   COMMON_DATA *cmonData, SBR_ELEMENT_TYPE sbrElem, int coupling)
{
  int payloadBits = 0;

  switch (sbrElem) {
  case SBR_ID_SCE:
    payloadBits += encodeSbrSingleChannelElement (sbrEnvDataLeft, &cmonData->sbrBitbuf);
    break;
  case SBR_ID_CPE:
    payloadBits += encodeSbrChannelPairElement (sbrEnvDataLeft, sbrEnvDataRight, &cmonData->sbrBitbuf, coupling);
    break;
  default:
    /* we never should apply SBR to any other element type */
    break;
  } 
  cmonData->sbrDataBits = payloadBits;
  return payloadBits;
}

/*****************************************************************************
    description:  writes SBR single channel data element
*****************************************************************************/
int WriteEnvSingleChannelElement(SBR_ENV_DATA *sbrEnvData, COMMON_DATA *cmonData)
{
  int payloadBits = 0;
  cmonData->sbrDataBits = 0;

  /* write pure sbr data */
  if (sbrEnvData != NULL) {
    /* write data */
    payloadBits += encodeSbrData (sbrEnvData, NULL, cmonData, SBR_ID_SCE, 0);
  }
  return payloadBits;
}

/*****************************************************************************
    description:  writes SBR channel pair data element
*****************************************************************************/
int WriteEnvChannelPairElement (SBR_ENV_DATA *sbrEnvDataLeft, SBR_ENV_DATA *sbrEnvDataRight,
								COMMON_DATA *cmonData, int coupling)
{
  int payloadBits = 0;
  cmonData->sbrDataBits = 0;

  /* write pure sbr data */
  if ((sbrEnvDataLeft != NULL) && (sbrEnvDataRight != NULL)) {
    /* write data */
    payloadBits += encodeSbrData (sbrEnvDataLeft, sbrEnvDataRight, cmonData, SBR_ID_CPE, coupling);
  }
  return payloadBits;
}

/*****************************************************************************
    description:  writes bits corresponding to the noise-floor-level
*****************************************************************************/
int writeNoiseLevelData (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream, int coupling)
{
  int j, i, payloadBits = 0;
  int nNoiseEnvelopes = ((sbrEnvData->noOfEnvelopes > 1) ? 2 : 1);

  for (i = 0; i < nNoiseEnvelopes; i++) {
    if (coupling && sbrEnvData->balance) {
      payloadBits += WriteBits (hBitStream,
                                sbrEnvData->sbr_noise_levels[i * sbrEnvData->noOfnoisebands],
                                sbrEnvData->NFHuff.start_bits_balance);
    } else {
      payloadBits += WriteBits (hBitStream,
                                sbrEnvData->sbr_noise_levels[i * sbrEnvData->noOfnoisebands],
                                sbrEnvData->NFHuff.start_bits);
    }
    for (j = 1 + i * sbrEnvData->noOfnoisebands; j < (sbrEnvData->noOfnoisebands * (1 + i)); j++) {
      if (coupling && sbrEnvData->balance) {
        payloadBits += WriteBits (hBitStream,
                                  bookSbrEnvBalanceC11F[sbrEnvData->sbr_noise_levels[j] +
                                                                          CODE_BOOK_SCF_LAV_BALANCE11],
                                  sbrEnvData->NFHuff.hufftableBalanceFreqL[sbrEnvData->sbr_noise_levels[j] +
                                                                          CODE_BOOK_SCF_LAV_BALANCE11]);
      } else {
        payloadBits += WriteBits (hBitStream,
                                  v_Huff_envelopeLevelC11F[sbrEnvData->sbr_noise_levels[j] +
                                                                 CODE_BOOK_SCF_LAV11],
                                  sbrEnvData->NFHuff.hufftableFreqL[sbrEnvData->sbr_noise_levels[j] +
                                                                 CODE_BOOK_SCF_LAV11]);
      }
    }
  }
  return payloadBits;
}


/*****************************************************************************
    description:  writes bits corresponding to the envelope data
*****************************************************************************/
int writeEnvelopeData (SBR_ENV_DATA *sbrEnvData, BIT_BUF *hBitStream, int coupling)
{
  int payloadBits = 0, j, i, delta;

  for (j = 0; j < sbrEnvData->noOfEnvelopes; j++) { /* loop over all envelopes */
    if (coupling && sbrEnvData->balance) {
      payloadBits += WriteBits (hBitStream, sbrEnvData->ienvelope[j][0], sbrEnvData->EnvHuff.start_bits_balance);
    } else {
      payloadBits += WriteBits (hBitStream, sbrEnvData->ienvelope[j][0], sbrEnvData->EnvHuff.start_bits);
    }

    for (i = 1 - sbrEnvData->domain_vec[j]; i < sbrEnvData->noScfBands[j]; i++) {
      delta = sbrEnvData->ienvelope[j][i];
      if (coupling && sbrEnvData->balance) {
          payloadBits += WriteBits (hBitStream,
                                    sbrEnvData->EnvHuff.hufftableBalanceFreqC[delta + sbrEnvData->EnvHuff.codeBookScfLavBalance],
                                    sbrEnvData->EnvHuff.hufftableBalanceFreqL[delta + sbrEnvData->EnvHuff.codeBookScfLavBalance]);
      } else {
        payloadBits += WriteBits (hBitStream,
                                  sbrEnvData->EnvHuff.hufftableFreqC[delta + sbrEnvData->EnvHuff.codeBookScfLav],
                                  sbrEnvData->EnvHuff.hufftableFreqL[delta + sbrEnvData->EnvHuff.codeBookScfLav]);
      }
    }
  }
  return payloadBits;
}

void InitSbrHuffmanTables (SBR_ENV_DATA *sbrEnvData, AMP_RES amp_res)
{
  sbrEnvData->init_sbr_amp_res = amp_res;

  switch (amp_res) {
  case  SBR_AMP_RES_3_0:
    sbrEnvData->EnvHuff.hufftableBalanceFreqC = bookSbrEnvBalanceC11F;
    sbrEnvData->EnvHuff.hufftableBalanceFreqL = bookSbrEnvBalanceL11F;
    sbrEnvData->EnvHuff.hufftableFreqC        = v_Huff_envelopeLevelC11F;
    sbrEnvData->EnvHuff.hufftableFreqL        = v_Huff_envelopeLevelL11F;
    sbrEnvData->EnvHuff.codeBookScfLavBalance = CODE_BOOK_SCF_LAV_BALANCE11;
    sbrEnvData->EnvHuff.codeBookScfLav = CODE_BOOK_SCF_LAV11;
    sbrEnvData->EnvHuff.start_bits = SI_SBR_START_ENV_BITS_AMP_RES_3_0;
    sbrEnvData->EnvHuff.start_bits_balance = SI_SBR_START_ENV_BITS_BALANCE_AMP_RES_3_0;
    break;

  case SBR_AMP_RES_1_5:
    sbrEnvData->EnvHuff.hufftableBalanceFreqC = bookSbrEnvBalanceC10F;
    sbrEnvData->EnvHuff.hufftableBalanceFreqL = bookSbrEnvBalanceL10F;
    sbrEnvData->EnvHuff.hufftableFreqC        = v_Huff_envelopeLevelC10F;
    sbrEnvData->EnvHuff.hufftableFreqL        = v_Huff_envelopeLevelL10F;
    sbrEnvData->EnvHuff.codeBookScfLavBalance = CODE_BOOK_SCF_LAV_BALANCE10;
    sbrEnvData->EnvHuff.codeBookScfLav = CODE_BOOK_SCF_LAV10;
    sbrEnvData->EnvHuff.start_bits = SI_SBR_START_ENV_BITS_AMP_RES_1_5;
    sbrEnvData->EnvHuff.start_bits_balance = SI_SBR_START_ENV_BITS_BALANCE_AMP_RES_1_5;
    break;

  default:
    break;
  }

  sbrEnvData->NFHuff.hufftableBalanceFreqL = bookSbrEnvBalanceL11F;
  sbrEnvData->NFHuff.hufftableFreqL        = v_Huff_envelopeLevelL11F;
  sbrEnvData->NFHuff.codeBookScfLavBalance = CODE_BOOK_SCF_LAV_BALANCE11;
  sbrEnvData->NFHuff.codeBookScfLav = CODE_BOOK_SCF_LAV11;

  sbrEnvData->NFHuff.start_bits         = SI_SBR_START_NOISE_BITS_AMP_RES_3_0;
  sbrEnvData->NFHuff.start_bits_balance = SI_SBR_START_NOISE_BITS_BALANCE_AMP_RES_3_0;

  sbrEnvData->EnvHuff.upDate = 0;
  sbrEnvData->NFHuff.upDate = 0;
}

