/*
  Envelope estimation
*/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "sbr.h"

float  sbr_envYBuffer[MAX_CHANNELS  * QMF_TIME_SLOTS * QMF_CHANNELS];
float  sbr_envRBuffer [MAX_CHANNELS][QMF_TIME_SLOTS][QMF_CHANNELS];
float  sbr_envIBuffer [MAX_CHANNELS][QMF_TIME_SLOTS][QMF_CHANNELS];

SBR_FRAME_INFO
frameInfo1_2048 = {1,
                   {0, 16},
                   1,
                   {0, 16}};

SBR_FRAME_INFO
frameInfo2_2048 = {2,
                   {0, 8, 16},
                   2,
                   {0, 8, 16}};

SBR_FRAME_INFO
frameInfo4_2048 = {4,
                   {0, 4, 8, 12, 16},
                   2,
                   {0, 8, 16}};

static inline SBR_FRAME_INFO *frameInfoGenerator(SBR_FRAME_INFO *hSbrFrameInfo, int nEnv)
{
  switch (nEnv) {
  case 1:
      memcpy (hSbrFrameInfo, &frameInfo1_2048, sizeof (SBR_FRAME_INFO));
      break;
  case 2:
      memcpy (hSbrFrameInfo, &frameInfo2_2048, sizeof (SBR_FRAME_INFO));
      break;
  case 4:
      memcpy (hSbrFrameInfo, &frameInfo4_2048, sizeof (SBR_FRAME_INFO));
      break;
  default:
    break;
  }
  return hSbrFrameInfo;
}

static inline void getEnergyFromCplxQmfData(float **energyValues, float realValues[32][64], float imagValues[32][64])
{
  int j, k;
  for (k = 0; k < 16; k++) {
    for (j = 0; j < 64; j++) {
      energyValues[k][j] = ((realValues[2*k][j] * realValues[2*k][j] +
                             imagValues[2*k][j] * imagValues[2*k][j]) +
                            (realValues[2*k+1][j] * realValues[2*k+1][j] +
                             imagValues[2*k+1][j] * imagValues[2*k+1][j])) * 0.5f;
    }
  }
}

#define QUANT_ERROR_THRES 200

/***************************************************************************
  \brief  Quantisation of the panorama value
  \return the quantized pan value
****************************************************************************/
static int mapPanorama (int nrgVal, int ampRes, int *quantError)
{
  int i;
  int min_val, val;
  int panTable[2][10] = {{0,2,4,6,8,12,16,20,24},
                         { 0, 2, 4, 8, 12 }};
  int maxIndex[2] = {9,5};

  int panIndex;
  int sign;
  sign = nrgVal > 0 ? 1 : -1;
  nrgVal = sign * nrgVal;
  min_val = INT_MAX;
  panIndex = 0;

  for (i = 0; i < maxIndex[ampRes]; i++) {
    val = abs (nrgVal - panTable[ampRes][i]);
    if (val < min_val) {
      min_val = val;
      panIndex = i;
    }
  }

  *quantError=min_val;
  return panTable[ampRes][maxIndex[ampRes]-1] + sign * panTable[ampRes][panIndex];
}

/***************************************************************************
  \brief  Quantisation of the noise floor levels
****************************************************************************/
static void sbrNoiseFloorLevelsQuantisation (int *iNoiseLevels, float *NoiseLevels, int coupling)
{
  int i;
  int dummy;
  for (i = 0; i < MAX_NUM_NOISE_VALUES; i++) {
    int tmp;
    tmp = NoiseLevels[i] > (float)30.0f ? 30: (int) (NoiseLevels[i] + (float)0.5);
    if (coupling) {
      tmp = tmp < -30 ? -30  : tmp;
      tmp = mapPanorama (tmp,1,&dummy);
    }
    iNoiseLevels[i] = tmp;
  }
}

static inline float fastpow2(float val) {
    typedef union {
        int i;
        float f;
    } conv;

    conv tmp;

    const float a = (float)((1 << 23));
    const float b = (float)((1 << 23) * 127);
    /* According to "A Fast, Compact Approximation of the Exponential Function"
     * by Nicol N. Schraudolph, 60801.48 yields the minimum RMS error for the
     * piecewise-linear approximation when using doubles (20 bits residual).
     * We have 23 bits, so we scale this value by 8. */
    const float c = 60801.48f * 8.f;
    tmp.i = (int)(a * val + (b - c + 0.5f));
    return tmp.f;
}

/***************************************************************************
  \brief  Calculation of noise floor for coupling
****************************************************************************/
static void coupleNoiseFloor(float *noise_level_left, float *noise_level_right)
{
  int i;
  for (i = 0; i < MAX_NUM_NOISE_VALUES; i++) {
    float temp1, temp2;
    //temp1 = powf(2.0f, (- noise_level_right[i] + NOISE_FLOOR_OFFSET));
    //temp2 = powf(2.0f, (- noise_level_left[i]  + NOISE_FLOOR_OFFSET));
    temp1 = fastpow2((- noise_level_right[i] + NOISE_FLOOR_OFFSET));
    temp2 = fastpow2((- noise_level_left[i]  + NOISE_FLOOR_OFFSET));
    noise_level_left[i] =  (NOISE_FLOOR_OFFSET - logf((temp1 + temp2) / 2) * I_LN2);
    noise_level_right[i] = (logf(temp2 / temp1) * I_LN2);
  }
}

static void mapLowResEnergyVal(int currVal,int* prevData, int index)
{
  prevData[index] = currVal;
}

static int computeBits (int delta, int codeBookScfLavLevel, int codeBookScfLavBalance,
                        const unsigned char * hufftableLevel, const unsigned char * hufftableBalance,
                        int coupling, int channel)
{
  int index;
  int delta_bits = 0;

  if (coupling && (channel == 1)) {
    index = (delta < 0) ? max (delta, -codeBookScfLavBalance) : min (delta, codeBookScfLavBalance);
    if (index != delta) {
      return (10000);
    }
    delta_bits = hufftableBalance[index + codeBookScfLavBalance];
  } else {
    index = (delta < 0) ? max (delta, -codeBookScfLavLevel) : min (delta, codeBookScfLavLevel);
    if (index != delta) {
      return (10000);
    }
    delta_bits = hufftableLevel[index + codeBookScfLavLevel];
  }
  return (delta_bits);
}

static void codeEnvelope (int *sfb_nrg, int nSfb, SBR_CODE_ENVELOPE * h_sbrCodeEnvelope, int *directionVec,
           int coupling, int nEnvelopes, int channel, int force_dF, int headerActive)
{
  int i, no_of_bands, band, last_nrg, curr_nrg;
  int *ptr_nrg;

  int codeBookScfLavLevel;
  int codeBookScfLavBalance;
  //const unsigned char *hufftableLevelTimeL;
  //const unsigned char *hufftableBalanceTimeL;
  const unsigned char *hufftableLevelFreqL;
  const unsigned char *hufftableBalanceFreqL;
  int envDataTableCompFactor;

  int delta_F_bits = 0, delta_T_bits = 0;
  int use_dT;

  int delta_F[MAX_FREQ_COEFFS];
  int delta_T[MAX_FREQ_COEFFS];

  float dF_edge_1stEnv = h_sbrCodeEnvelope->dF_edge_1stEnv +
    h_sbrCodeEnvelope->dF_edge_incr * h_sbrCodeEnvelope->dF_edge_incr_fac;

  codeBookScfLavLevel = h_sbrCodeEnvelope->codeBookScfLav;
  //hufftableLevelTimeL = h_sbrCodeEnvelope->hufftableTimeL;
  hufftableLevelFreqL = h_sbrCodeEnvelope->hufftableFreqL;
  if (coupling) {
    codeBookScfLavBalance = h_sbrCodeEnvelope->codeBookScfLavBalance;
   // hufftableBalanceTimeL = h_sbrCodeEnvelope->hufftableBalanceTimeL;
    hufftableBalanceFreqL = h_sbrCodeEnvelope->hufftableBalanceFreqL;
  } else {
    codeBookScfLavBalance = h_sbrCodeEnvelope->codeBookScfLav;
    //hufftableBalanceTimeL = h_sbrCodeEnvelope->hufftableTimeL;
    hufftableBalanceFreqL = h_sbrCodeEnvelope->hufftableFreqL;
  }

  if(coupling == 1 && channel == 1)
    envDataTableCompFactor = 1;
  else
    envDataTableCompFactor = 0;

  if (headerActive) {
    h_sbrCodeEnvelope->upDate = 0;
  }

  for (i = 0; i < nEnvelopes; i++) {
    no_of_bands = nSfb;
    ptr_nrg = sfb_nrg;
    curr_nrg = *ptr_nrg;

    delta_F[0] = curr_nrg >> envDataTableCompFactor;
    if (coupling && channel == 1)
      delta_F_bits = h_sbrCodeEnvelope->start_bits_balance;
    else
      delta_F_bits = h_sbrCodeEnvelope->start_bits;

    /* if(h_sbrCodeEnvelope->upDate != 0) {
      delta_T[0] = (curr_nrg - h_sbrCodeEnvelope->sfb_nrg_prev[0]) >> envDataTableCompFactor;
      delta_T_bits = computeBits (delta_T[0], codeBookScfLavLevel, codeBookScfLavBalance,
                                  hufftableLevelTimeL, hufftableBalanceTimeL, coupling, channel);
    } */

    mapLowResEnergyVal(curr_nrg, h_sbrCodeEnvelope->sfb_nrg_prev, 0);

    /* ensure that nrg difference is not higher than codeBookScfLavXXXFreq */
    if (coupling && channel == 1) {
      for (band = no_of_bands - 1; band > 0; band--) {
        if (ptr_nrg[band] - ptr_nrg[band-1] > codeBookScfLavBalance) {
          ptr_nrg[band-1] = ptr_nrg[band] - codeBookScfLavBalance;
        }
      }
      for (band = 1; band < no_of_bands; band++) {
        if (ptr_nrg[band-1] - ptr_nrg[band] > codeBookScfLavBalance) {
          ptr_nrg[band] = ptr_nrg[band-1] - codeBookScfLavBalance;
        }
      }
    } else {
      for (band = no_of_bands - 1; band > 0; band--) {
        if (ptr_nrg[band] - ptr_nrg[band-1] > codeBookScfLavLevel) {
          ptr_nrg[band-1] = ptr_nrg[band] - codeBookScfLavLevel;
        }
      }
      for (band = 1; band < no_of_bands; band++) {
        if (ptr_nrg[band-1] - ptr_nrg[band] > codeBookScfLavLevel) {
          ptr_nrg[band] = ptr_nrg[band-1] - codeBookScfLavLevel;
        }
      }
    }

    for (band = 1; band < no_of_bands; band++) {
      last_nrg = (*ptr_nrg);
      ptr_nrg++;
      curr_nrg = (*ptr_nrg);
      delta_F[band] = (curr_nrg - last_nrg) >> envDataTableCompFactor;
      delta_F_bits += computeBits (delta_F[band], codeBookScfLavLevel, codeBookScfLavBalance, hufftableLevelFreqL,
                                   hufftableBalanceFreqL, coupling, channel);

      /* if(h_sbrCodeEnvelope->upDate != 0) {
        delta_T[band] = curr_nrg - h_sbrCodeEnvelope->sfb_nrg_prev[band];
        delta_T[band] = delta_T[band] >> envDataTableCompFactor;
      } */

      mapLowResEnergyVal(curr_nrg, h_sbrCodeEnvelope->sfb_nrg_prev, band);

      /* if(h_sbrCodeEnvelope->upDate != 0) {
        delta_T_bits += computeBits (delta_T[band], codeBookScfLavLevel, codeBookScfLavBalance,
                                     hufftableLevelTimeL, hufftableBalanceTimeL, coupling, channel);
      } */
    }

    /* if (i == 0) {
      use_dT = (h_sbrCodeEnvelope->upDate != 0 && (delta_F_bits > delta_T_bits * (1 + dF_edge_1stEnv)));
    } else {
      use_dT = (delta_F_bits > delta_T_bits);
    } */

    //if (use_dT) {
      //directionVec[i] = TIME;
      //memcpy (sfb_nrg, delta_T, no_of_bands * sizeof (int));
    //} else {
      directionVec[i] = FREQ;
      memcpy (sfb_nrg, delta_F, no_of_bands * sizeof (int));
    //}

    sfb_nrg += no_of_bands;
    h_sbrCodeEnvelope->upDate = 1;
  }
}

/***************************************************************************
  \brief  calculates the SBR envelope values
****************************************************************************/
static void calculateSbrEnvelope (float **YBufferLeft, float **YBufferRight, const SBR_FRAME_INFO *frame_info,
				  int *sfb_nrgLeft, int *sfb_nrgRight, SBR_CONFIG_DATA *h_con, ENV_CHANNEL *h_sbr,
				  SBR_STEREO_MODE stereoMode, int* maxQuantError)
{
  int i, j, k, l, count, m = 0;
  int no_of_bands, start_pos, stop_pos, li, ui;

  int ca = 2 - h_sbr->encEnvData.init_sbr_amp_res;
  int quantError;
  int nEnvelopes = frame_info->nEnvelopes;
  int short_env = - 1;
  int timeStep = 2;
  int missingHarmonic;

  if (stereoMode == SBR_COUPLING) {
    *maxQuantError = 0;
  }

  for (i = 0; i < nEnvelopes; i++) {
    start_pos = timeStep * frame_info->borders[i];
    stop_pos = timeStep * frame_info->borders[i + 1];
    no_of_bands = h_con->nSfb[FREQ_RES_HIGH];

    if (i == short_env) {
      stop_pos = stop_pos - timeStep;
    }

    for (j = 0; j < no_of_bands; j++) {
      float nrgRight = 0;
      float nrgLeft = 0;
      float temp;

      li = h_con->freqBandTable[FREQ_RES_HIGH][j];
      ui = h_con->freqBandTable[FREQ_RES_HIGH][j + 1];

      if(j == 0 && ui-li > 1){
        li++;
      }

      missingHarmonic = 0;
      if(h_sbr->encEnvData.addHarmonicFlag){
        if(h_sbr->encEnvData.addHarmonic[j]){
          missingHarmonic = 1;
        }
      }

      if(missingHarmonic){
        float tmpNrg = 0;
        count = (stop_pos - start_pos);
        for (l = start_pos; l < stop_pos; l++) {
          nrgLeft += YBufferLeft[l/2][li];
        }
        
        for (k = li+1; k < ui; k++){
          tmpNrg = 0;
          for (l = start_pos; l < stop_pos; l++) {
            tmpNrg += YBufferLeft[l/2][k];
          }
          if(tmpNrg > nrgLeft){
            nrgLeft = tmpNrg;
          }
        }
        
        if(ui-li > 2){
          nrgLeft = nrgLeft*0.398107267f;
        } else{
          if(ui-li > 1){
            nrgLeft = nrgLeft*0.5f;
          }
        }
        
        if (stereoMode == SBR_COUPLING) {
          for (l = start_pos; l < stop_pos; l++) {
            nrgRight += YBufferRight[l/2][li];
          }
          for (k = li+1; k < ui; k++){
            tmpNrg = 0;
            for (l = start_pos; l < stop_pos; l++) {
              tmpNrg += YBufferRight[l/2][k];
            }
            if(tmpNrg > nrgRight){
              nrgRight = tmpNrg;
            }
          }
          if(ui-li > 2){
            nrgRight = nrgRight*0.398107267f;
          } else{
            if(ui-li > 1){
              nrgRight = nrgRight*0.5f;
            }
          }
        }
        if (stereoMode == SBR_COUPLING) {
          temp = nrgLeft;
          nrgLeft = (nrgRight + nrgLeft) * (float)0.5f;
          nrgRight = (temp + 1) / (nrgRight + 1);
        }
    } else { /* end missingHarmonic */
     count = (stop_pos - start_pos) * (ui - li);
     for (k = li; k < ui; k++) {
        for (l = start_pos; l < stop_pos; l++) {
          nrgLeft += YBufferLeft[l/2][k];
        }
      }
      if (stereoMode == SBR_COUPLING) {
        for (k = li; k < ui; k++) {
          for (l = start_pos; l < stop_pos; l++) {
            nrgRight += YBufferRight[l/2][k];
          }
        }
        temp = nrgLeft;
        nrgLeft = (nrgRight + nrgLeft) * 0.5f;
        nrgRight = (temp + 1) / (nrgRight + 1);
      }
	}
      nrgLeft = logf(nrgLeft / (count * 64) + EPS)*I_LN2;
      nrgLeft = max(nrgLeft,0);
      sfb_nrgLeft[m] = (int) (ca * nrgLeft + 0.5);
      if (stereoMode == SBR_COUPLING) {
        nrgRight = logf(nrgRight) * I_LN2;
        sfb_nrgRight[m] = mapPanorama ((int)((float)ca*nrgRight),h_sbr->encEnvData.init_sbr_amp_res,&quantError);
        if(quantError > *maxQuantError) {
          *maxQuantError = quantError;
        }
      }
      m++;
    } /* j */

    if (h_con->useParametricCoding) {
      m-=no_of_bands;
      for (j = 0; j < no_of_bands; j++) {
        if(sfb_nrgLeft[m] < 0) {
          sfb_nrgLeft[m] = 0;
        }
        m++;
      }
    }
  } /* i*/
}

typedef struct {
  float  r00r;
  float  r11r;
  float  r01r;
  float  r01i;
  float  r02r;
  float  r02i;
  float  r12r;
  float  r12i;
  float  r22r;
} ACORR_COEFS;

void calcAutoCorrSecondOrder (ACORR_COEFS *ac, float realBuf[32][64], float imagBuf[32][64], int bd)
{
  int   j, jminus1, jminus2;

  memset(ac, 0, sizeof(ACORR_COEFS));
  for (j = 0; j < 14 - 1; j++) {
    float a,b,c,d,e,f;
    jminus1 = j - 1;
    jminus2 = jminus1 - 1;
    a = realBuf[j][bd];
    b = realBuf[jminus1][bd];
    c = realBuf[jminus2][bd];
    d = imagBuf[j][bd];
    e = imagBuf[jminus1][bd];
    f = imagBuf[jminus2][bd];
    ac->r00r += a*a + d*d;
    ac->r11r += b*b + e*e;
    ac->r01r += a*b + d*e;
    ac->r01i += d*b - a*e;
    ac->r02r += a*c + d*f;
    ac->r02i += d*c - a*f;
  }

  ac->r22r = ac->r11r + realBuf[-2][bd] * realBuf[-2][bd] +
                        imagBuf[-2][bd] * imagBuf[-2][bd];
  ac->r12r = ac->r01r + realBuf[-1][bd] * realBuf[-2][bd] +
                        imagBuf[-1][bd] * imagBuf[-2][bd];
  ac->r12i = ac->r01i + imagBuf[-1][bd] * realBuf[-2][bd] -
                        realBuf[-1][bd] * imagBuf[-2][bd];

  jminus1 = j - 1;
  jminus2 = jminus1 - 1;

  ac->r00r += realBuf[j][bd] * realBuf[j][bd] +
              imagBuf[j][bd] * imagBuf[j][bd];
  ac->r11r += realBuf[jminus1][bd] * realBuf[jminus1][bd] +
              imagBuf[jminus1][bd] * imagBuf[jminus1][bd];
  ac->r01r += realBuf[j][bd] * realBuf[jminus1][bd] +
              imagBuf[j][bd] * imagBuf[jminus1][bd];
  ac->r01i += imagBuf[j][bd] * realBuf[jminus1][bd] -
              realBuf[j][bd] * imagBuf[jminus1][bd];
  ac->r02r += realBuf[j][bd] * realBuf[jminus2][bd] +
              imagBuf[j][bd] * imagBuf[jminus2][bd];
  ac->r02i += imagBuf[j][bd] * realBuf[jminus2][bd] -
              realBuf[j][bd] * imagBuf[jminus2][bd];
}

/**************************************************************************
  \brief Calculates the tonal to noise ration
**************************************************************************/
void CalculateTonalityQuotas(SBR_TON_CORR_EST *hTonCorr, float sourceBufferReal[32][64], float sourceBufferImag[32][64], int usb)
{
  int    i, k, r, timeIndex;
  float  alphar[2], alphai[2], det;
  ACORR_COEFS ac;
  float*  nrgVector         = hTonCorr->nrgVector;
  float** quotaMatrix       = hTonCorr->quotaMatrix;

  /* Buffering of the quotaMatrix and the quotaMatrixTransp. */
  for(i =  0 ; i < 2; i++){
    memcpy(quotaMatrix[i],
           quotaMatrix[i + 2],
           QMF_CHANNELS * sizeof(float));
  }

  memmove(nrgVector,nrgVector+2,2*sizeof(float));
  memset(nrgVector+2,0,2*sizeof(float));

  for (r = 0; r < usb; r++) {
    k = 2;
    timeIndex = 2;
    while(k <= 32 - 14) {
      calcAutoCorrSecondOrder (&ac, &sourceBufferReal[k], &sourceBufferImag[k], r);
      det = ac.r11r * ac.r22r - (ac.r12r * ac.r12r + ac.r12i * ac.r12i);

      if (det == 0 ) {
        alphar[1] = alphai[1] = 0;
      } else {
        alphar[1] = ( ac.r01r * ac.r12r - ac.r01i * ac.r12i - ac.r02r * ac.r11r ) / det;
        alphai[1] = ( ac.r01i * ac.r12r + ac.r01r * ac.r12i - ac.r02i * ac.r11r ) / det;
      }

      if (ac.r11r == 0) {
        alphar[0] =  alphai[0] = 0;
      } else {
        alphar[0] = - ( ac.r01r + alphar[1] * ac.r12r + alphai[1] * ac.r12i ) / ac.r11r;
        alphai[0] = - ( ac.r01i + alphai[1] * ac.r12r - alphar[1] * ac.r12i ) / ac.r11r;
      }

      if(ac.r00r){
        float tmp;
        tmp =  -(alphar[0]*ac.r01r + alphai[0]*ac.r01i + alphar[1]*ac.r02r + alphai[1]*ac.r02i)/(ac.r00r);
        quotaMatrix[timeIndex][r] = tmp / ((float)1.0 - tmp +(float)RELAXATION);
      } else {
        quotaMatrix[timeIndex][r] = 0;
      }

      nrgVector[timeIndex] += ac.r00r;
      k += 16;
      timeIndex++;
    }
  }
}

/***************************************************************************
  \brief  calculates the noise floor and the envelope values.
****************************************************************************/
void extractSbrEnvelope (float *timeInPtr, float *pCoreBuffer, unsigned int timeInStride, SBR_CONFIG_DATA *h_con, SBR_HEADER_DATA *sbrHeaderData,
			 SBR_BITSTREAM_DATA *sbrBitstreamData, ENV_CHANNEL *h_envChan[MAX_CHANNELS], COMMON_DATA *hCmonData)
{
  int ch, i, j, c;
  int nEnvelopes[MAX_CHANNELS];
  const SBR_FRAME_INFO *frame_info[MAX_CHANNELS];
  int nChannels = h_con->nChannels;
  int nInChannels = nChannels;
  SBR_STEREO_MODE stereoMode = h_con->stereoMode;
  int   sfb_nrg    [MAX_CHANNELS][MAX_NUM_ENVELOPE_VALUES];
  float noiseFloor[MAX_CHANNELS][MAX_NUM_NOISE_VALUES];
  int   noise_level[MAX_CHANNELS][MAX_NUM_NOISE_VALUES];
  int   maxQuantError;
  memset(noiseFloor,0,sizeof(noiseFloor));

  for(ch = 0; ch < nInChannels;ch++){
      sbrAnalysisFiltering  (&sbrHeaderData->fft32, timeInPtr ? timeInPtr+ch:NULL, timeInStride, sbr_envRBuffer[ch], sbr_envIBuffer[ch],
                             h_envChan[ch]->qmf_states_buffer);
  } /* ch */

  for(ch = 0; ch < nChannels;ch++) {
    getEnergyFromCplxQmfData (h_envChan[ch]->YBuffer + 16,
                              sbr_envRBuffer[ch], sbr_envIBuffer[ch]);

    CalculateTonalityQuotas(&h_envChan[ch]->TonCorr, sbr_envRBuffer[ch], sbr_envIBuffer[ch],
                            h_con->freqBandTable[HI][h_con->nSfb[HI]]);
  } /* ch */
  frame_info[0] = frameInfoGenerator (&h_envChan[0]->SbrEnvFrame, h_envChan[0]->encEnvData.bs_num_env);

  switch (stereoMode) {
  case SBR_LEFT_RIGHT:
    frame_info[1] = frameInfoGenerator (&h_envChan[1]->SbrEnvFrame, h_envChan[0]->encEnvData.bs_num_env);
    break;

  case SBR_COUPLING:
    frame_info[1] = frame_info[0];
    break;

  case SBR_MONO:
    /* nothing to do */
    break;

  default:
    break;
  }

  for(ch = 0; ch < nChannels;ch++){
    h_envChan[ch]->encEnvData.noOfEnvelopes = nEnvelopes[ch] = frame_info[ch]->nEnvelopes;

    for (i = 0; i < nEnvelopes[ch]; i++){
      h_envChan[ch]->encEnvData.noScfBands[i] = h_con->nSfb[FREQ_RES_HIGH];
    }

    TonCorrParamExtr(&h_envChan[ch]->TonCorr, h_envChan[ch]->encEnvData.sbr_invf_mode_vec, noiseFloor[ch],
                     &h_envChan[ch]->encEnvData.addHarmonicFlag, h_envChan[ch]->encEnvData.addHarmonic,
                     frame_info[ch]->nNoiseEnvelopes, h_con->freqBandTable[HI], h_con->nSfb[HI]);

    h_envChan[ch]->encEnvData.sbr_invf_mode = h_envChan[ch]->encEnvData.sbr_invf_mode_vec[0];
    h_envChan[ch]->encEnvData.noOfnoisebands = h_envChan[ch]->TonCorr.sbrNoiseFloorEstimate.noNoiseBands;
  } /* ch */

  switch (stereoMode) {
  case SBR_MONO:
    calculateSbrEnvelope (h_envChan[0]->YBuffer, NULL, frame_info[0], sfb_nrg[0],
                          NULL,h_con,h_envChan[0], SBR_MONO,NULL);
    break;

  case SBR_LEFT_RIGHT:
    calculateSbrEnvelope (h_envChan[0]->YBuffer, NULL, frame_info[0], sfb_nrg[0],
                          NULL, h_con, h_envChan[0], SBR_MONO,NULL);
    calculateSbrEnvelope (h_envChan[1]->YBuffer, NULL, frame_info[1], sfb_nrg[1],
                          NULL, h_con,h_envChan[1], SBR_MONO,NULL);
    break;

  case SBR_COUPLING:
    calculateSbrEnvelope (h_envChan[0]->YBuffer, h_envChan[1]->YBuffer, frame_info[0],
                          sfb_nrg[0], sfb_nrg[1], h_con,h_envChan[0], SBR_COUPLING,&maxQuantError);
    break;

  default:
    break;
  }

  switch (stereoMode) {
  case SBR_MONO:
    sbrNoiseFloorLevelsQuantisation (noise_level[0], noiseFloor[0], 0);
    codeEnvelope (noise_level[0], h_envChan[0]->encEnvData.noOfnoisebands,
                  &h_envChan[0]->encEnvData.NFHuff,
                  h_envChan[0]->encEnvData.domain_vec_noise, 0,
                  (frame_info[0]->nEnvelopes > 1 ? 2 : 1), 0, 0,
                 sbrBitstreamData->HeaderActive);
    break;

  case SBR_LEFT_RIGHT:
    sbrNoiseFloorLevelsQuantisation (noise_level[0],noiseFloor[0], 0);
    codeEnvelope (noise_level[0], h_envChan[0]->encEnvData.noOfnoisebands,
                  &h_envChan[0]->encEnvData.NFHuff,
                  h_envChan[0]->encEnvData.domain_vec_noise, 0,
                  (frame_info[0]->nEnvelopes > 1 ? 2 : 1), 0, 0,
                  sbrBitstreamData->HeaderActive);
    sbrNoiseFloorLevelsQuantisation (noise_level[1],noiseFloor[1], 0);
    codeEnvelope (noise_level[1], h_envChan[1]->encEnvData.noOfnoisebands,
                  &h_envChan[0]->encEnvData.NFHuff,
                  h_envChan[1]->encEnvData.domain_vec_noise, 0,
                  (frame_info[1]->nEnvelopes > 1 ? 2 : 1), 0, 0,
                  sbrBitstreamData->HeaderActive);
    break;

  case SBR_COUPLING:
    coupleNoiseFloor(noiseFloor[0],noiseFloor[1]);
    sbrNoiseFloorLevelsQuantisation (noise_level[0],noiseFloor[0], 0);
    codeEnvelope (noise_level[0], h_envChan[0]->encEnvData.noOfnoisebands,
                  &h_envChan[0]->encEnvData.NFHuff,
                  h_envChan[0]->encEnvData.domain_vec_noise, 1,
                  (frame_info[0]->nEnvelopes > 1 ? 2 : 1), 0, 0,
                  sbrBitstreamData->HeaderActive);
    sbrNoiseFloorLevelsQuantisation (noise_level[1],noiseFloor[1], 1);
    codeEnvelope (noise_level[1], h_envChan[1]->encEnvData.noOfnoisebands,
                  &h_envChan[0]->encEnvData.NFHuff,
                  h_envChan[1]->encEnvData.domain_vec_noise, 1,
                  (frame_info[1]->nEnvelopes > 1 ? 2 : 1), 1, 0,
                  sbrBitstreamData->HeaderActive);
    break;

  default:
    break;
  }

  switch (stereoMode) {
  case SBR_MONO:
    sbrHeaderData->coupling = 0;
    h_envChan[0]->encEnvData.balance = 0;
    codeEnvelope (sfb_nrg[0], h_con->nSfb[FREQ_RES_HIGH],
                  &h_envChan[0]->encEnvData.EnvHuff,
                  h_envChan[0]->encEnvData.domain_vec,
                  sbrHeaderData->coupling,
                  frame_info[0]->nEnvelopes, 0, 1,
                  sbrBitstreamData->HeaderActive);
    break;

  case SBR_LEFT_RIGHT:
    sbrHeaderData->coupling = 0;
    h_envChan[0]->encEnvData.balance = 0;
    h_envChan[1]->encEnvData.balance = 0;
    codeEnvelope (sfb_nrg[0], h_con->nSfb[FREQ_RES_HIGH],
                  &h_envChan[0]->encEnvData.EnvHuff,
                  h_envChan[0]->encEnvData.domain_vec,
                  sbrHeaderData->coupling,
                  frame_info[0]->nEnvelopes, 0, 1,
                  sbrBitstreamData->HeaderActive);
    codeEnvelope (sfb_nrg[1], h_con->nSfb[FREQ_RES_HIGH],
                  &h_envChan[0]->encEnvData.EnvHuff,
                  h_envChan[1]->encEnvData.domain_vec,
                  sbrHeaderData->coupling,
                  frame_info[1]->nEnvelopes, 0, 1,
                  sbrBitstreamData->HeaderActive);
    break;

  case SBR_COUPLING:
    sbrHeaderData->coupling = 1;
    h_envChan[0]->encEnvData.balance = 0;
    h_envChan[1]->encEnvData.balance = 1;
    codeEnvelope (sfb_nrg[0], h_con->nSfb[FREQ_RES_HIGH],
                  &h_envChan[0]->encEnvData.EnvHuff,
                  h_envChan[0]->encEnvData.domain_vec,
                  sbrHeaderData->coupling,
                  frame_info[0]->nEnvelopes, 0, 1,
                  sbrBitstreamData->HeaderActive);
    codeEnvelope (sfb_nrg[1], h_con->nSfb[FREQ_RES_HIGH],
                  &h_envChan[0]->encEnvData.EnvHuff,
                  h_envChan[1]->encEnvData.domain_vec,
                  sbrHeaderData->coupling,
                  frame_info[1]->nEnvelopes, 1, 1, 
                  sbrBitstreamData->HeaderActive);
    break;

  default:
    break;
  } /* end switch(stereoMode) */

  if (nChannels == 1) {
    if (h_envChan[0]->encEnvData.domain_vec[0] == TIME) {
      h_envChan[0]->encEnvData.EnvHuff.dF_edge_incr_fac++;
    } else {
      h_envChan[0]->encEnvData.EnvHuff.dF_edge_incr_fac = 0;
    }
  } else {
    if (h_envChan[0]->encEnvData.domain_vec[0] == TIME ||
        h_envChan[1]->encEnvData.domain_vec[0] == TIME) {
      h_envChan[0]->encEnvData.EnvHuff.dF_edge_incr_fac++;
      h_envChan[1]->encEnvData.EnvHuff.dF_edge_incr_fac++;
    } else {
      h_envChan[0]->encEnvData.EnvHuff.dF_edge_incr_fac = 0;
      h_envChan[1]->encEnvData.EnvHuff.dF_edge_incr_fac = 0;
    }
  }

  for(ch = 0; ch < nChannels;ch++){
    c = 0;
    for (i = 0; i < nEnvelopes[ch]; i++) {
      for (j = 0; j < h_envChan[ch]->encEnvData.noScfBands[i]; j++) {
        h_envChan[ch]->encEnvData.ienvelope[i][j] = sfb_nrg[ch][c];
        c++;
      }
    }
    for (i = 0; i < MAX_NUM_NOISE_VALUES; i++){
      h_envChan[ch]->encEnvData.sbr_noise_levels[i] = noise_level[ch][i];
    }
  }/* ch */

  if (nChannels == 2) {
    WriteEnvChannelPairElement(&h_envChan[0]->encEnvData, &h_envChan[1]->encEnvData,
                               hCmonData, sbrHeaderData->coupling);
  } else {
    WriteEnvSingleChannelElement(&h_envChan[0]->encEnvData, hCmonData);
  }

  for(ch = 0; ch < nChannels;ch++){
    for (i = 0; i < 16; i++) {
      float *temp;
      temp = h_envChan[ch]->YBuffer[i];
      h_envChan[ch]->YBuffer[i] = h_envChan[ch]->YBuffer[i+16];
      h_envChan[ch]->YBuffer[i+16] = temp;
    }
  }/* ch */
  sbrHeaderData->prev_coupling = sbrHeaderData->coupling;
}

