/*
  Noise floor estimation
*/
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ton_corr.h"

const float smoothFilter[4]  = {0.05857864376269f, 0.2f, 0.34142135623731f, 0.4f};

static void smoothingOfNoiseLevels (float *NoiseLevels, int nEnvelopes, int noNoiseBands,
                             		float prevNoiseLevels[MAX_NUM_NOISE_VALUES][NF_SMOOTHING_LENGTH])
{
  int i,band,env;
  for(env = 0; env < nEnvelopes; env++){
    for (i = 0; i < noNoiseBands; i++){
      prevNoiseLevels[i][0] = prevNoiseLevels[i][1];
      prevNoiseLevels[i][1] = prevNoiseLevels[i][2];
      prevNoiseLevels[i][2] = prevNoiseLevels[i][3];
	  prevNoiseLevels[i][3] = *(NoiseLevels + env*noNoiseBands + i);
    }
    for (band = 0; band < noNoiseBands; band++){
      NoiseLevels[band+ env*noNoiseBands] = 0;
      for (i = 0; i < NF_SMOOTHING_LENGTH; i++){
        NoiseLevels[band+env*noNoiseBands] += smoothFilter[i]*prevNoiseLevels[band][i];
      }
    }
  }
}

/**************************************************************************
  \brief     Does the noise floor level estiamtion.
**************************************************************************/
static void qmfBasedNoiseFloorDetection(float *noiseLevel, float** quotaMatrixOrig, char* indexVector,
                                 		int startIndex, int stopIndex, int startChannel, int stopChannel,
                                 		float ana_max_level, float noiseFloorOffset, float weightFac,
								 		INVF_MODE diffThres, INVF_MODE inverseFilteringLevel)
{
  int l,k;
  float meanOrig=0, meanSbr=0, diff;
  float tonalityOrig, tonalitySbr;

  for(l = startChannel; l < stopChannel;l++){
    tonalityOrig = 0;
    for(k = startIndex ; k < stopIndex; k++){
      tonalityOrig += quotaMatrixOrig[k][l];
    }
    tonalityOrig /= (stopIndex-startIndex);
    tonalitySbr = 0;
    for(k = startIndex ; k < stopIndex; k++){
      tonalitySbr += quotaMatrixOrig[k][indexVector[l]];
    }
    tonalitySbr /= (stopIndex-startIndex);
    meanOrig += tonalityOrig;
    meanSbr += tonalitySbr;
  }
  meanOrig /= (stopChannel - startChannel);
  meanSbr /= (stopChannel - startChannel);
  if(meanOrig < (float)0.000976562f && meanSbr < (float)0.000976562f){
    meanOrig = 101.5936673f;
    meanSbr  = 101.5936673f;
  }
  if(meanOrig < (float)1.0f) {
    meanOrig = 1.0f;
  }
  if(meanSbr < (float)1.0f) {
    meanSbr  = 1.0f;
  }
  diff = max ((float)1.0f, weightFac*meanSbr/meanOrig);
  if(inverseFilteringLevel == INVF_MID_LEVEL ||
     inverseFilteringLevel == INVF_LOW_LEVEL ||
     inverseFilteringLevel == INVF_OFF){
    diff = 1.0f;
  }
  if (inverseFilteringLevel <= diffThres) {
    diff = 1.0f;
  }
  *noiseLevel =  diff/meanOrig;
  *noiseLevel *= noiseFloorOffset;
  *noiseLevel = min (*noiseLevel, (float)ana_max_level);
}

/**************************************************************************
  \brief     Does the noise floor level estiamtion..
**************************************************************************/
void NoiseFloorEstimateQmf (SBR_NOISE_FLOOR_ESTIMATE *h_sbrNoiseFloorEstimate,
						    int nNoiseEnvelopes,
                            float *noiseLevels,
                            float** quotaMatrixOrig,
                            char* indexVector,
                            INVF_MODE* pInvFiltLevels)
{
  int startPos[2], stopPos[2], env, band;
  int noNoiseBands      = h_sbrNoiseFloorEstimate->noNoiseBands;
  int *freqBandTable    = h_sbrNoiseFloorEstimate->freqBandTableQmf;

  if(nNoiseEnvelopes == 1){
    startPos[0] = 0;
    stopPos[0]  = 2;
  } else{
    startPos[0] = 0;
    stopPos[0]  = 1;
    startPos[1] = 1;
    stopPos[1]  = 2;
  }

  for(env = 0; env < nNoiseEnvelopes; env++){
    for(band = 0; band < noNoiseBands; band++){
      qmfBasedNoiseFloorDetection(&noiseLevels[band + env*noNoiseBands],
                                  quotaMatrixOrig, indexVector,
                                  startPos[env], stopPos[env],
                                  freqBandTable[band], freqBandTable[band+1],
                                  h_sbrNoiseFloorEstimate->ana_max_level,
                                  h_sbrNoiseFloorEstimate->noiseFloorOffset[band],
                                  h_sbrNoiseFloorEstimate->weightFac,
                                  h_sbrNoiseFloorEstimate->diffThres,
                                  pInvFiltLevels[band]);

    }
  }

  smoothingOfNoiseLevels(noiseLevels, nNoiseEnvelopes, h_sbrNoiseFloorEstimate->noNoiseBands,
                         h_sbrNoiseFloorEstimate->prevNoiseLevels);

  for(env = 0; env < nNoiseEnvelopes; env++){
    for(band = 0; band < noNoiseBands; band++){
      noiseLevels[band + env*noNoiseBands] = NOISE_FLOOR_OFFSET - (I_LN2*logf(noiseLevels[band + env*noNoiseBands]));
    }
  }
}

static int downSampleLoRes(int *v_result, int num_result, const unsigned char *freqBandTableRef, int num_Ref)
{
  int step;
  int i,j;
  int org_length,result_length;
  int v_index[MAX_FREQ_COEFFS/2];
  org_length=num_Ref;
  result_length=num_result;
  v_index[0]=0;	/* Always use left border */
  i=0;
  while(org_length > 0) {
      i++;
      step=org_length/result_length;
      org_length=org_length - step;
      result_length--;
      v_index[i]=v_index[i-1]+step;
  }
  if(i != num_result) {
    return (1);/* error */
  }
  for(j=0;j<=i;j++) {
      v_result[j]=freqBandTableRef[v_index[j]];
  }
  return (0);
}

/**************************************************************************
  \brief     Resets the current instance of the noise floor estiamtion module.
**************************************************************************/
static int resetSbrNoiseFloorEstimate (SBR_NOISE_FLOOR_ESTIMATE *h_sbrNoiseFloorEstimate,
                                	   const unsigned char *freqBandTable, int nSfb)
{
  int k2,kx;
  k2=freqBandTable[nSfb];
  kx=freqBandTable[0];
  if(h_sbrNoiseFloorEstimate->noiseBands == 0){
    h_sbrNoiseFloorEstimate->noNoiseBands = 1;
  } else{
    h_sbrNoiseFloorEstimate->noNoiseBands =(int) ((h_sbrNoiseFloorEstimate->noiseBands * logf((float)k2/kx) * I_LN2)+0.5);
    if(h_sbrNoiseFloorEstimate->noNoiseBands==0) {
      h_sbrNoiseFloorEstimate->noNoiseBands=1;
    }
  }
  return(downSampleLoRes(h_sbrNoiseFloorEstimate->freqBandTableQmf, h_sbrNoiseFloorEstimate->noNoiseBands, freqBandTable,nSfb));
}

/**************************************************************************
  \brief    Creates an instance of the noise floor level estimation module.
**************************************************************************/
int CreateSbrNoiseFloorEstimate (SBR_NOISE_FLOOR_ESTIMATE  *h_sbrNoiseFloorEstimate, int ana_max_level,
                             	 const unsigned char *freqBandTable, int nSfb, int noiseBands,
				 				 unsigned int useSpeechConfig)
{
  int i;
  memset(h_sbrNoiseFloorEstimate,0,sizeof(SBR_NOISE_FLOOR_ESTIMATE));

  if (useSpeechConfig) {
    h_sbrNoiseFloorEstimate->weightFac = 1.0;
    h_sbrNoiseFloorEstimate->diffThres = INVF_LOW_LEVEL;
  } else {
    h_sbrNoiseFloorEstimate->weightFac = 0.25;
    h_sbrNoiseFloorEstimate->diffThres = INVF_MID_LEVEL;
  }

  h_sbrNoiseFloorEstimate->ana_max_level          = 0.5f; //(float) pow(2,ana_max_level/3.0f);
  h_sbrNoiseFloorEstimate->noiseBands             = noiseBands;

  if(resetSbrNoiseFloorEstimate(h_sbrNoiseFloorEstimate,freqBandTable,nSfb)) {
    return(1);
  }

  for(i=0;i<h_sbrNoiseFloorEstimate->noNoiseBands;i++) {
    h_sbrNoiseFloorEstimate->noiseFloorOffset[i] = 1.0f;
  }

  return 0;
}

